#!/usr/bin/env python

"""
kio_adfs.py

IOSlave for ADFS disk images.
Use adfs:<URL> to view the contents of an image at <URL>.

Copyright (C) 2003-2005 David Boddie <david@boddie.org.uk>

This software is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of
the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public
License along with this library; see the file COPYING
If not, write to the Free Software Foundation, Inc.,
59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

"""

from qt import QByteArray
from kio import KIO
from kdecore import KURL
import os, select, sys, traceback, time

# Import the ADFSlib module.

import ADFSlib

# Debugging variable and functions --------------------------------------------

# Debugging flag: set to True to enable debugging.

DEBUG = False
DEBUG_FILE = "/tmp/kio_debug.txt"

def debug_file():

    return open(DEBUG_FILE, "a")

def debug(msg):

    if not DEBUG: return
    
    debug_file().write(msg+"\n")
    sys.stderr.write(msg+"\n")

# Create a trace method to catch errors in methods.

def trace(fn, args, file = None):

    debug("tracing %s%s" % (fn.__name__, args))
    try:
    
        result = fn(*args)
        debug("exiting %s with %s\n" % (fn.__name__, repr(result)))
        return result
    
    except:
    
        if file is None:
            file = open(DEBUG_FILE, "a")
        
        traceback.print_exception(
            *(sys.exc_info()), **{"file": file}
            )
    
    debug("exiting %s\n" % fn.__name__)

# The trap function allows functions to be called under conditions where
# any exceptions raised are caught and written to the debug file.

def trap(fn):

    """fn = trap(fn)
    
    Create a function wrapper in which the original function is called
    within a try...except clause for debugging purposes.
    """

    old_fn = fn
    return lambda *args: trace(old_fn, args)

# End of debugging functions --------------------------------------------------


# Define a class which will be used to create IOSlave instances.

class SlaveClass(KIO.SlaveBase):

    """SlaveClass(KIO.SlaveBase)
    
    See kdelibs/kio/kio/slavebase.h for virtual functions to override.
    """
    
    def __init__(self, pool, app):
    
        # We must call the initialisation method of the base class.
        
        KIO.SlaveBase.__init__(self, "adfs", pool, app)
        
        # Initialise various instance variables.
        
        self.host = ""
        self.disc_image = None
        self.adfsdisc = None
        
        self.readers = \
        {
            0xfca00:    self.squash_reader,
            0xddc00:    self.nspark_reader
        }
        
        self.info_handlers = \
        {
            0xfca00:    self.squash_info,
            0xddc00:    self.nspark_info
        }
    
    def parse_url(self, url):
    
        file_path = unicode(url.path())
        
        at = file_path.find(u":")
        
        if at != -1:
        
            file_path = file_path[at+1:]
        
        # If we have cached a URL for the disc image then check whether
        # this URL refers to an item beneath it. This implies that the
        # cached URL is a substring of the one given.
        
        if self.disc_image and file_path.find(self.disc_image) == 0:
        
            # Return the path within the image.
            return file_path[len(self.disc_image):]
        
        # Descend the URL, element by element, as far as possible.
        elements = file_path.split(u"/")
        
        if len(elements) < 2:
        
            return None
        
        path_elements, elements = elements[:1], elements[1:]
        
        while elements != []:
        
            path_elements.append(elements.pop(0))
            
            path = u"/".join(path_elements)
            
            if os.path.isfile(path):
            
                break
            
            elif elements == [] and os.path.isdir(path):
            
                return None
            
            elif not os.path.exists(path):
            
                return None
        
        # The image is found at the constructed path.
        self.disc_image = path
        
        # The characters following this path correspond to the path within
        # the image file.
        image_path = u"/".join(elements)
        
        # If not already open, it is necessary to open the image file.
        try:
        
            adf = open(self.disc_image, "rb")
        
        except IOError:
        
            return None
        
        try:
        
            self.adfsdisc = ADFSlib.ADFSdisc(adf)
        
        except ADFSlib.ADFS_exception:
        
            adf.close()
            return None
        
        return image_path
    
    def get(self, url):
    
        path = self.parse_url(url)
        
        if path is None:
        
            self.error(KIO.ERR_DOES_NOT_EXIST, url.path())
            return
        
        # Find the object within the image.
        adfs_object = self.find_file_within_image(path)
        
        if not adfs_object:
        
            self.error(KIO.ERR_DOES_NOT_EXIST, path)
            return
        
        # Check whether it is a directory and return an error if so.
        if type(adfs_object[1]) == type([]):
        
            self.error(KIO.ERR_IS_DIRECTORY, path)
            return
        
        #self.mimeType(self.image_mimetype)
        
        # If the object is a file then the list returned will take the form:
        # [name, data, load, exec, length]
        # and if we are using filetypes then we can append a suffix derived
        # from the load address of (load >> 8) & 0xfff
        
        data = adfs_object[1]
        
        if self.adfsdisc.disc_type.find("adE") != -1:
        
            reader = self.readers.get(adfs_object[2] & 0xfff00, None)
            
            if reader:
            
                data = reader(data)
                
                if data is None:
                
                    self.error(KIO.ERR_IS_DIRECTORY, path)
                    return
        
        self.data(QByteArray(data))
        
        # The end of the data string.
        self.data(QByteArray())
        
        self.finished()
    
    #get = trap(get)
    
    def stat(self, url):
    
        path = self.parse_url(url)
        
        if path is None:
        
            # Try redirecting to the protocol contained in the path.
            redir_url = KURL(url.path())
            self.redirection(redir_url)
            self.finished()
            #self.error(KIO.ERR_DOES_NOT_EXIST, url.path())
            return
        
        # Find the object within the image.
        adfs_object = self.find_file_within_image(path)
        
        if not adfs_object:
        
            self.error(KIO.ERR_DOES_NOT_EXIST, path)
            return
        
        if type(adfs_object[1]) != type([]):
        
            entry = self.build_entry(adfs_object)
        
        elif path != u"" and path[-1] != u"/":
        
            # Directory referenced, but URL does not end in a slash.
            
            url.setPath(unicode(url.path()) + u"/")
            self.redirection(url)
            self.finished()
            return
        
        else:
        
            entry = self.build_entry(adfs_object)
        
        if entry != []:
        
            # Return the entry to the application, indicating that it is
            # the only/last entry.
            self.statEntry(entry)
            
            self.finished()
        
        else:
        
            self.error(KIO.ERR_DOES_NOT_EXIST, path)
    
    #stat = trap(stat)
    
#    def mimetype(self, url):
#    
#        self.mimeType("text/plain")
#        self.finished()
#        #debug("mimetype: %s" % unicode(url))
#        
#        name = self.parse_url(url)
#        
#        if name is None:
#        
#            self.error(KIO.ERR_DOES_NOT_EXIST, url.path())
#            return
#        
#        # Return the default image MIME type.
#        self.mimeType(self.image_mimetype)
#        
#        self.finished()
#    
#    mimetype = trap(mimetype)
    
    def listDir(self, url):
    
        # The "url" argument is a kdecore.KURL object.
        
        path = self.parse_url(url)
        
        if path is None:
        
            # Try redirecting to the protocol contained in the path.
            redir_url = KURL(url.path())
            
            print redir_url.url()
            
            self.redirection(redir_url)
            self.finished()
            #self.error(KIO.ERR_DOES_NOT_EXIST, url.path())
            return
        
        # Find the object within the image.
        adfs_object = self.find_file_within_image(path)
        
        if not adfs_object:
        
            self.error(KIO.ERR_DOES_NOT_EXIST, path)
            return
        
        elif path != u"" and path[-1] != u"/":
        
            # The referenced object may a directory, but the URL does not
            # end in a slash so redirect to ensure that a directory is
            # being referenced. This will either result in a directory
            # being found or an error being returned to the client.
            
            url.setPath(unicode(url.path()) + u"/")
            self.redirection(url)
            self.finished()
            return
        
        elif type(adfs_object[1]) != type([]):
        
            self.error(KIO.ERR_IS_FILE, path)
            return
        
        # Obtain a list of files.
        files = adfs_object[1]
        
        # Return the objects in the list to the application.
        
        for this_file in files:
        
            #debug("name: %s" % name)
            #debug("attr: %s" % repr(attributes))
            
            entry = self.build_entry(this_file)
            
            if entry != []:
            
                self.listEntry(entry, 0)
                
                # For old style disk images, return a .inf file, too.
                if self.adfsdisc.disc_type.find("adE") == -1:
                
                    this_inf = self.find_file_within_image(
                        path + "/" + this_file[0] + ".inf"
                        )
                    
                    if this_inf is not None:
                    
                        entry = self.build_entry(this_inf)
                        
                        if entry != []:
                        
                            self.listEntry(entry, 0)
        
        # We have finished listing the contents of a directory.
        self.listEntry([], 1)
        
        self.finished()
    
    #listDir = trap(listDir)
    
    #def disconnectSlave(self):
    #
    #    #debug("disconnectSlave")
    #    pass
    
    def dispatchLoop(self):
    
        #debug("dispatchLoop")
        KIO.SlaveBase.dispatchLoop(self)
    
    def build_entry(self, obj):
    
        entry = []
        
        # Check the type of object passed.
        if type(obj[1]) != type([]):
        
            # [name, data, load, exec, length]
            name = self.encode_name_from_object(obj)
            
            if self.adfsdisc.disc_type.find("adE") == -1 and \
                obj[0][-4:] == ".inf":
            
                # For .inf files, use a MIME type of text/plain.
                mimetype = "text/plain"
                length = obj[4]
            
            else:
            
                info_handler = self.info_handlers.get(obj[2] & 0xfff00, None)
                
                # Provide default values for the file type, MIME type and
                # length.
                filetype = None
                mimetype = None
                length = obj[4]
                
                if info_handler:
                
                    filetype, mimetype, length = info_handler(
                        obj, filetype, mimetype, length
                        )
        
        else:
        
            name = self.encode_name_from_object(obj)
            length = 0
            mimetype = "inode/directory"
        
        #debug("name: %s" % name)
        
        atom = KIO.UDSAtom()
        atom.m_uds = KIO.UDS_NAME
        atom.m_str = name
        
        entry.append(atom)
        
        if length:
        
            atom = KIO.UDSAtom()
            atom.m_uds = KIO.UDS_SIZE
            atom.m_long = length
            
            entry.append(atom)
        
        atom = KIO.UDSAtom()
        atom.m_uds = KIO.UDS_MODIFICATION_TIME
        # Number of seconds since the epoch.
        atom.m_long = int(time.time())
        
        entry.append(atom)
        
        atom = KIO.UDSAtom()
        atom.m_uds = KIO.UDS_ACCESS
        # The usual octal permission information (rw-r--r-- in this case).
        atom.m_long = 0644
        
        entry.append(atom)
        
        # If the stat method is implemented then entries _must_ include
        # the UDS_FILE_TYPE atom or the whole system may not work at all.
        
        atom = KIO.UDSAtom()
        atom.m_uds = KIO.UDS_FILE_TYPE
        
        if mimetype != "inode/directory":
        
            atom.m_long = os.path.stat.S_IFREG
        
        else:
        
            atom.m_long = os.path.stat.S_IFDIR
        
        entry.append(atom)
        
        if mimetype:
        
            atom = KIO.UDSAtom()
            atom.m_uds = KIO.UDS_MIME_TYPE
            atom.m_str = mimetype
            
            entry.append(atom)
        
        return entry
    
    def encode_name_from_object(self, obj):
    
        name = obj[0]
        
        # If the name contains a slash then replace it with a dot.
        new_name = u".".join(name.split(u"/"))
        
        if self.adfsdisc.disc_type.find("adE") == 0:
        
            if type(obj[1]) != type([]) and u"." not in new_name:
            
                # Construct a suffix from the object's load address/filetype.
                
                info_handler = self.info_handlers.get(obj[2] & 0xfff00, None)
                
                # Provide default values for the file type, MIME type and
                # length.
                filetype = (obj[2] >> 8) & 0xfff
                mimetype = None
                length = obj[4]
                
                if info_handler:
                
                    filetype, mimetype, length = info_handler(
                        obj, filetype, mimetype, length
                        )
                
                suffix = u"%03x" % filetype
                new_name = new_name + "." + suffix
        
        return unicode(KURL.encode_string_no_slash(new_name))
    
    def decode_name(self, name):
    
        return unicode(KURL.decode_string(name))
    
    def find_file_within_image(self, path, objs = None):
    
        if objs is None:
        
            objs = self.adfsdisc.files
        
        elements = path.split(u"/")
        
        # Remove any empty elements.
        elements = filter(lambda x: x != u"", elements)
        
        if elements == []:
        
            # Special case for root directory.
            return [u"/", objs, 0, 0, 0]
        
        for this_obj in objs:
        
            # Examine each object and compare its name to the next path
            # element.
            
            if type(this_obj[1]) != type([]):
            
                # A file is found. 
                obj_name = self.encode_name_from_object(this_obj)
                
                if obj_name == elements[0]:
                
                    # A match between names.
                    
                    if len(elements) == 1:
                    
                        # This is the last path element; we have found the
                        # required file.
                        return this_obj
                    
                    else:
                    
                        # There are more elements to satisfy but we can
                        # descend no further.
                        return None
                
                elif self.adfsdisc.disc_type.find("adE") == -1 and \
                     elements[0] == obj_name + u".inf":
                
                    # Old style discs will have .inf files, too.
                    
                    if len(elements) == 1:
                    
                        # Construct a .inf file to return to the client.
                        file_data = "%s\t%X\t%X\t%X\n" % \
                            tuple(this_obj[:1] + this_obj[2:])
                        
                        new_obj = \
                        (
                            this_obj[0] + ".inf", file_data,
                            0, 0, len(file_data)
                        )
                        
                        return new_obj
                    
                    else:
                    
                        # There are more elements to satisfy but we can
                        # descend no further.
                        return None
            
            else:
            
                # A directory is found.
                obj_name = self.encode_name_from_object(this_obj)
                
                if obj_name == elements[0]:
                
                    # A match between names.
                    
                    if len(elements) == 1:
                    
                        # This is the last path element; we have found the
                        # required file.
                        return this_obj
                    
                    else:
                    
                        # More path elements need to be satisfied; descend
                        # further.
                        return self.find_file_within_image(
                            u"/".join(elements[1:]), this_obj[1]
                            )
        
        # No matching objects were found.
        return None
    
    def remote_process(self, command, data):
    
        si, so, se = os.popen3(command)
        output = []
        
        length = len(data)
        
        total = 0
        ptr = 0
        
        while not (so.closed and si.closed):
        
            if not so.closed:
                reading = [so]
            else:
                reading = []
            
            if not si.closed:
                writing = [si]
            else:
                writing = []
            
            if not se.closed:
                others = [se]
            else:
                others = []
            
            r, w, x = select.select(reading, writing, others)
            
            if x != []:
            
                # Data on stderr
                return None
            
            elif w != []:
            
                # stdin available for writing
                if ptr < length:
                
                    w[0].write(data[ptr:ptr+1024])
                    ptr = ptr + 1024
                
                else:
                
                    w[0].close()
            
            if r != []:
            
                # Data available on stdout
                retrieved = r[0].read(1024)
                
                total = total + len(retrieved)
                
                if len(retrieved) == 0:
                
                    r[0].close()
                
                else:
                
                    output.append(retrieved)
            
            elif r == [] and w == []:
            
                # No input or output streams available.
                return None
        
        return "".join(output)
    
    def squash_reader(self, data):
    
        # A Squash file: use gunzip to decompress the data.
        command = "gunzip -c -"
        
        # Write the object's data (without the header) to gunzip's
        # stdin.
        return self.remote_process(command, data[20:])
    
    def squash_info(self, obj, def_filetype, def_mimetype, def_length):
    
        # Each Squash file contains the length of the data they
        # contain once it is decompressed.
        
        try:
        
            length = self.adfsdisc.str2num(4, obj[1][4:8])
        
        except IndexError:
        
            # Use the default length supplied.
            length = def_length
        
        # For Squash files, use the filetype in the header for
        # this file's suffix. (Use the disc image's str2num
        # method to extract the value from the file.)
        
        try:
        
            filetype = (
                self.adfsdisc.str2num(4, obj[1][8:12]) >> 8
                ) & 0xfff
        
        except IndexError:
        
            filetype = 0xfca
        
        # Let the client discover the MIME type by reading
        # the file.
        mimetype = None
        
        return filetype, mimetype, length
    
    def nspark_reader(self, data):
    
        # Archives which are handled by nspark are not read like ordinary
        # files, so return None to indicate failure.
        
        # For now, return the data.
        return data
    
    def nspark_info(self, obj, def_filetype, def_mimetype, def_length):
    
        # For the archive as a whole, we run the nspark utility with
        # a contradictory set of options.
        
        #command = "nspark -qtv"
        
        # For now, just return the default values.
        return def_filetype, def_mimetype, def_length
