#!/usr/bin/env python

"""
kio_bookmarks.py

An IOSlave for viewing Konqueror's bookmarks.xml file.

Copyright (C) 2003-2005 David Boddie <david@boddie.org.uk>

This software is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of
the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public
License along with this library; see the file COPYING
If not, write to the Free Software Foundation, Inc.,
59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

"""

# For debugging purposes, import the sys and traceback modules.

import sys, traceback

# Debugging variable and functions --------------------------------------------

# Debugging flag: set to True to enable debugging.

DEBUG = False
DEBUG_FILE = "/tmp/logs/pydebug.txt"

def debug_file():

    return open(DEBUG_FILE, "a")

def debug(msg):

    if not DEBUG: return
    
    debug_file().write(msg+"\n")
    sys.stderr.write(msg+"\n")

# Create a trace method to catch errors in methods.

def trace(fn, args, file = None):

    debug("tracing %s%s" % (fn.__name__, args))
    try:
    
        result = fn(*args)
        debug("exiting %s with %s\n" % (fn.__name__, repr(result)))
        return result
    
    except:
    
        if file is None:
            file = open(DEBUG_FILE, "a")
        
        traceback.print_exception(
            *(sys.exc_info()), **{"file": file}
            )
    
    debug("exiting %s\n" % fn.__name__)

# The trap function allows functions to be called under conditions where
# any exceptions raised are caught and written to the debug file.

def trap(fn):

    """fn = trap(fn)
    
    Create a function wrapper in which the original function is called
    within a try...except clause for debugging purposes.
    """

    old_fn = fn
    return lambda *args: trace(old_fn, args)

# End of debugging functions --------------------------------------------------

# Import the required Qt and KDE modules.

from qt import QFile, QString, QByteArray, QDataStream, IO_ReadOnly
from kio import KIO
from kdecore import KLockFile, KURL, locateLocal

# The os and time libraries are required for this IOSlave to function.

import os, time

# Compatibility hack. Versions of Python earlier than 2.2 didn't have this
# string in the os module.

if not os.__dict__.has_key("extsep"):

    os.extsep = "."

# The Qt XML library is used to read and interpret the bookmarks.xml file.

import qtxml


# Define a class which will be used to create IOSlave instances.

class SlaveClass(KIO.SlaveBase):

    """SlaveClass(KIO.SlaveBase)
    
    See kdelibs/kio/kio/slavebase.h for virtual functions to override.
    """

# Earlier versions of this IOSlave represented bookmarks as HTML files
# containing a description of the bookmark details.
    page_template = \
u"""
<html>
  <head>
    <title>%(title)s</title>
  </head>
  <body>
    <h1>%(title)s</h1>
    <p>
      <img src="%(icon)s"><a href="%(href)s">%(title)s</a> (%(href)s)
    </p>
  </body>
</html>
"""

# The current version of this IOSlave represents bookmarks as Desktop files.    
    desktop_template = \
u"""[Desktop Entry]
Encoding=%(encoding)s
Icon=%(icon)s
Name=%(title)s
Type=Link
URL=%(href)s
"""
    
    # Use of "application/desktop" with files containing "forbidden"
    # characters for URLs such as "/" and "%", even when quoted
    # properly caused the self.mimeType call in get() to fail.
    bookmark_mimetype = "application/x-desktop"
    
    # Some possible alternative MIME types:
    #bookmark_mimetype = "application/desktop"
    #bookmark_mimetype = "application/x-webarchive"
    
    def __init__(self, pool, app):
    
        # We must call the initialisation method of the base class.
        
        KIO.SlaveBase.__init__(self, "bookmarks", pool, app)
        
        # Attach the DCOP client object associated with this IOSlave to the
        # DCOP server.
        self.dcopClient().attach()
        
        #debug("dcopClient: %i" % self.dcopClient().isRegistered())
        
        # Initialise some instance variables.
        self.host = ""
        self.document = None
        self.file = None
        
        debug("Exiting __init__ now");
    
    #__init__ = trap(__init__)
    
    def __del__(self):
    
        pass
    
    def setHost(self, host, port, user, passwd):
    
        #debug(
        #    "setHost: %s %s %s %s" % (
        #        repr(unicode(host)), repr(unicode(port)),
        #        repr(unicode(user)), repr(unicode(passwd))
        #        )
        #    )
        
        # This IOSlave does not allow a host to be specified as part of
        # a URL.
        
        if unicode(host) != u"":
        
            self.closeConnection()
            self.error(KIO.ERR_MALFORMED_URL, host)
            return
        
        #self.host = unicode(host)
        #self.finished()
    
    #setHost = trap(setHost)
    
    #def setSubURL(self, url):
    #
    #    #debug("setSubURL")
    #    self.finished()
    
    #setSubURL = trap(setSubURL)
    
    def openConnection(self):
    
        # Don't call self.finished() in this method.
        #debug("openConnection")
        
        # Find the user's home directory.
        self.home = os.getenv("HOME")
        
        # Find the bookmarks.xml file beneath the user's home directory.
        path = os.path.join(
            self.home, ".kde", "share", "apps", "konqueror",
            "bookmarks" + os.extsep + "xml"
            )
        
        self.temp_file = locateLocal("tmp", "kio_bookmarks")
        self.lock_file = KLockFile(self.temp_file)
        
        # Wait for the file to be unlocked.
        while self.lock_file.lock() != KLockFile.LockOK:
            pass
        
        # Record the path used to obtain the bookmarks.
        self.file = path
        
        try:
        
            b = open(self.file, "r").read()
        
        except IOError:
        
            self.error(KIO.ERR_CANNOT_OPEN_FOR_READING, self.file)
            return
        
        # Convert the XML to a DOM representation.
        self.document = qtxml.QDomDocument()
        
        # Read the file into the document, converting the XML into a
        # DOM representation.
        
        result, errorMsg, errorLine, errorColumn = self.document.setContent(b)
        
        if result == 0:
        
            self.error(KIO.ERR_CANNOT_OPEN_FOR_READING, self.file)
            self.document = None
            return
        
        #debug("%s" % str(self.document.toString()))
    
    #openConnection = trap(openConnection)
    
    def _flush(self):
    
        # Internal function for synchronising the filesystem with the
        # bookmarks file.
        if self.document is None:
        
            return
        
        try:
        
            b = open(self.file, "w")
        
        except IOError:
        
            self.error(KIO.ERR_CANNOT_OPEN_FOR_WRITING, self.file)
            return
        
        # Write the XML to the bookmarks file.
        b.write(unicode(self.document.toString()).encode("UTF_8"))
        
        b.close()
        
        # Unset the document attribute to force the bookmarks file to be
        # re-read.
        self.document = None
    
    #_flush = trap(_flush)
    
    def closeConnection(self):
    
        # Don't call self.finished() in this method.
        #debug("closeConnection")
        
        # Delete the temporary file. If something else is waiting for it
        # then its lock will create a new lock file anyway.
        QFile(self.temp_file).remove()
        
        # This should be unnecessary.
        self.lock_file.unlock()
        
        self._flush()
        
        self.document = None
    
    #closeConnection = trap(closeConnection)
    
    def error(self, *args):
    
        self.closeConnection()
        KIO.SlaveBase.error(self, *args)
    
    def finished(self):
    
        self.closeConnection()
        KIO.SlaveBase.finished(self)
    
    def get(self, url):
    
        self.openConnection()
        
        # Parse the URL using the URL's methods.
        path, name, obj = self.find_object_from_url(url)
        
        if obj is None:
        
            #debug("%s does not exist" % path)
            self.error(KIO.ERR_DOES_NOT_EXIST, path)
            return
        
        # The obj variable is a document node. Check whether it is a
        # folder node and return an error if so.
        if unicode(obj.nodeName()) != u"bookmark":
        
            #debug("%s is a directory" % path)
            self.error(KIO.ERR_IS_DIRECTORY, path)
            return
        
        # Indicate the MIME type of the file.
        self.mimeType(self.bookmark_mimetype)
        
        # For bookmark nodes, return a Desktop file containing any icon
        # and link information.
        details = self.read_bookmark_details(obj)
        
        # Decode the title string so that it is presented in a more
        # conventional manner.
        details[u"title"] = self.decode_name(details[u"title"])
        
        text = self.desktop_template % details
        output = text.encode(details[u"encoding"])
        
        #debug("total size: %s" % repr(len(output)))
        #self.totalSize(long(len(output)))
        
        self.data(QByteArray(output))
        
        # The end of the data string.
        self.data(QByteArray())
        
        #debug("%s" % output.encode(details["encoding"]))
        
        self.finished()
    
    #get = trap(get)
    
    def put(self, url, permissions, overwrite, resume):
    
        #debug("put")
        
        self.openConnection()
        
        # Parse the URL using the URL's methods.
        path, name, obj = self.find_object_from_url(url)
        
        #debug("obj: %s" % repr(obj))
        #debug("overwrite: %s" % repr(overwrite))
        
        if (not overwrite) and (obj is not None):
        
            self.error(KIO.ERR_COULD_NOT_WRITE, path)
            return
        
        elif obj is not None:
        
            # Find the parent object of the existing object.
            parent_obj = obj.parentNode()
            
            # Remove the existing object.
            parent_obj.removeChild(obj)
            #parent_obj.unlink()
        
        else:
        
            # Find the parent-to-be.
            parent_path, parent_name, parent_obj = \
                self.find_parent_from_path(path)
            
            #debug("parent: %s" % repr(parent_obj))
            
            if parent_obj is None:
            
                self.error(KIO.ERR_COULD_NOT_WRITE, path)
                return
        
        # Find the destination name of the object to be created.
        try:
        
            dest_name = unicode(path).split(u"/")[-1]
        
        except IndexError:
        
            self.error(KIO.ERR_COULD_NOT_WRITE, path)
            return
        
        # Read data from the application.
        
        bytearray = QByteArray()
        bytes = 0
        
        data = ""
        
        while 1:
        
            self.dataReq()
            result = self.readData(bytearray)
            
            if result <= 0:
            
                # An error or the end of data was encountered.
                break
            
            # The number of bytes read is given in the result.
            bytes = bytes + result
            data = data + unicode(bytearray)
        
        if result < 0:
        
            self.error(KIO.ERR_CANNOT_OPEN_FOR_WRITING, path)
            return
        
        # The file passed should contain bookmark information encoded in
        # a similar form to the way we present the information to the user.
        # However, we are only interested in "href" and "icon" information
        # and not the title as this is defined by the destination URL.
        href, icon, title = self.extract_bookmark_info(data)
        
        if href is None:
        
            self.error(KIO.ERR_COULD_NOT_WRITE, path)
            return
        
        if icon is None:
        
            icon = u"www"
        
        if title is None:
        
            title = href
        
        # Create a child node of the parent.
        self.create_child(
            u"bookmark", {u"href": href, u"icon": icon, u"title": title},
            dest_name, parent_obj
            )
        
        # Synchronise the filesystem with the bookmarks file.
        self._flush()
        
        self.finished()
    
    #put = trap(put)
    
    def stat(self, url):
    
        #debug("stat: %s" % url.url(0,0))
        
        self.openConnection()
        
        # Parse the URL using the URL's methods.
        path, name, obj = self.find_object_from_url(url)
        
        #debug("obj: %s" % repr(obj))
        #debug("obj.nodeName(): %s" % str(obj.nodeName()))
        
        if obj is None:
        
            self.error(KIO.ERR_DOES_NOT_EXIST, path)
            return
        
        if unicode(obj.nodeName()) == u"folder" or \
            unicode(obj.nodeName()) == u"xbel":
        
            if path == u"" or path[-1] != u"/":
            
                url.setPath(path + u"/")
                self.redirection(url)
                self.finished()
                return
        
        entry = self.build_entry(obj)
        
        #debug("entry: %s" % repr(entry))
        
        if entry != []:
        
            # Return the entry to the application, indicating that it is
            # the only/last entry.
            self.statEntry(entry)
            
            self.finished()
        
        else:
        
            self.error(KIO.ERR_DOES_NOT_EXIST, path)
    
    #stat = trap(stat)
    
    def mimetype(self, url):
    
        #debug("mimetype: %s" % unicode(url))
        self.openConnection()
        
        # Parse the URL using the URL's methods.
        path, name, obj = self.find_object_from_url(url)
        
        if obj is None:
        
            self.error(KIO.ERR_DOES_NOT_EXIST, path)
            return
        
        if unicode(obj.nodeName()) == u"folder":
        
            self.mimeType("inode/directory")
        
        elif unicode(obj.nodeName()) == u"bookmark":
        
            self.mimeType(self.bookmark_mimetype)
        
        else:
        
            self.error(KIO.ERR_DOES_NOT_EXIST, path)
            return
        
        self.finished()
    
    #mimetype = trap(mimetype)
    
    def listDir(self, url):
    
        # The "url" argument is a kdecore.KURL object.
        
        #debug("listDir: %s" % str(url.prettyURL(0)))
        self.openConnection()
        
        # Parse the URL using the URL's methods.
        path, name, obj = self.find_object_from_url(url)
        
        if obj is None:
        
            self.error(KIO.ERR_DOES_NOT_EXIST, path)
            return
        
        if path == u"" or path[-1] != u"/":
        
            url.setPath(path + u"/")
            self.redirection(url)
            self.finished()
            return
        
        if unicode(obj.nodeName()) != u"folder" and \
            unicode(obj.nodeName()) != u"xbel":
        
            self.error(KIO.ERR_IS_FILE, path)
            return
        
        # Obtain a list of nodes.
        #files = self.strip_text_nodes(obj.childNodes)
        
        # Return the objects in the list of child nodes to the application.
        child = obj.firstChild()
        
        while not child.isNull():
        
            #debug("name: %s" % child.nodeName())
            
            entry = self.build_entry(child)
            
            #debug("entry: %s" % entry)
            
            if entry != []:
            
                self.listEntry(entry, 0)
            
            child = child.nextSibling()
        
        # We have finished listing the contents of a directory.
        self.listEntry([], 1)
        
        self.finished()
    
    #listDir = trap(listDir)
    
    def mkdir(self, url, permissions):
    
        #debug("mkdir")
        
        self.openConnection()
        
        # Parse the URL using the URL's methods.
        host, path = unicode(url.host()), unicode(url.path())
        
        # Find the potential parent-to-be of the new directory.
        parent_path, parent_name, obj = self.find_parent_from_path(path)
        
        if obj is None:
        
            self.error(KIO.ERR_DOES_NOT_EXIST, parent_path)
            return
        
        if unicode(obj.nodeName()) != u"folder" and \
            unicode(obj.nodeName()) != u"xbel":
        
            self.error(KIO.ERR_IS_FILE, parent_path)
            return
        
        # Find the destination object's name.
        try:
        
            dest_name = path.split(u"/")[-1]
        
        except IndexError:
        
            self.error(KIO.ERR_COULD_NOT_MKDIR, path)
            return
        
        # Create the directory in the node returned.
        
        if self.find_named_object(obj, dest_name) is not None:
        
            self.error(KIO.ERR_COULD_NOT_MKDIR, path)
            return
        
        # Add a new <folder> element to the parent node.
        self.create_child(
            u"folder", {u"folded": u"yes", u"icon": u"bookmark_folder"},
            dest_name, obj
            )
        
        # Synchronise the filesystem with the bookmarks file.
        self._flush()
        
        self.finished()
    
    #mkdir = trap(mkdir)
    
    def rename(self, src, dest, overwrite):
    
        #debug("rename: %s %s" % (src.path(), dest.path()))
        self.openConnection()
        
        # Find the source leafname and object from the URL given.
        src_path, src_name, src_obj = self.find_object_from_url(src)
        
        if src_obj is None:
        
            self.error(KIO.ERR_DOES_NOT_EXIST, src.path())
            return
        
        # Find the destination leafname and object from the URL given.
        dest_path, dest_name, dest_obj = self.find_object_from_url(dest)
        
        #debug("%s %s" % (repr(src_obj), repr(dest_obj)))
        
        if src_obj is dest_obj:
        
            #debug("No change")
            self.finished()
        
        if (not overwrite) and (dest_obj is not None):
        
            self.error(KIO.ERR_CANNOT_RENAME, dest.path())
            return
        
        elif overwrite and (dest_obj is not None):
        
            # Overwrite an existing destination object.
            parent_obj = dest_obj.parentNode()
            parent_obj.removeChild(dest_obj)
            #parent_obj.unlink()
        
        # Split the source and destination paths into elements and check
        # whether the locations exist within the same folder.
        src_leading = src_path.split(u"/")[:-1]
        dest_elements = dest_path.split(u"/")
        
        dest_leading = dest_elements[:-1]
        
        try:
        
            dest_name = dest_elements[-1]
        
        except IndexError:
        
            self.error(KIO.ERR_CANNOT_RENAME, dest_path)
            return
        
        # Change the title element within the source object. The title
        # element must exist because it was used to locate the object.
        
        if src_leading == dest_leading:
        
            # Change the name of the source object.
            self.retitle_object(src_obj, dest_name)
        
        else:
        
            # The destination URL is elsewhere within the directory
            # structure. Create a deep copy of the node, placing it at
            # the new location then remove the old node.
            #obj_copy = src_obj.cloneNode(1)
            
            # Remove the original node from its parent's list of children.
            src_parent = src_obj.parentNode()
            obj_copy = src_parent.removeChild(src_obj)
            
            # Find the parent-to-be of the destination node.
            parent_path, parent_name, parent_obj = \
                self.find_parent_from_path(dest_path)
            
            if parent_obj is None:
            
                self.error(KIO.ERR_DOES_NOT_EXIST, parent_path)
                return
            
            if unicode(parent_obj.nodeName()) != u"folder" and \
                unicode(parent_obj.nodeName()) != u"xbel":
            
                self.error(KIO.ERR_IS_FILE, parent_path)
                return
            
            # Add the new node to the parent's children.
            obj_copy = parent_obj.appendChild(obj_copy)
            
            if src_name != dest_name:
            
                # Change the name of the new object.
                self.retitle_object(obj_copy, dest_name)
        
        # Synchronise the filesystem with the bookmarks file.
        self._flush()
        
        self.finished()
    
    #rename = trap(rename)
    
    # Other possible file operations are represented by the following
    # methods which are not implemented.
    
    #def symlink(self, target, dest, overwrite):
    #
    #    debug("symlink")
    #    if self.document is None:
    #    
    #        self.openConnection()
    #    
    #    self.finished()
    
    #def chmod(self, url, permissions):
    #
    #    debug("chmod")
    #    if self.document is None:
    #    
    #        self.openConnection()
    #    
    #    self.finished()
    
    def copy(self, src, dest, permissions, overwrite):
    
        #debug("copy")
        self.openConnection()
        
        # Find the source path, name and object from the URL given.
        src_path, src_name, src_obj = self.find_object_from_url(src)
        
        if src_obj is None:
        
            self.error(KIO.ERR_DOES_NOT_EXIST, src.path())
            return
        
        # Find the destination path, name and object from the URL given.
        dest_path, dest_name, dest_obj = self.find_object_from_url(dest)
        
        if dest_obj is not None:
        
            # Object already exists
            
            if unicode(dest_obj.nodeName()) == u"folder":
            
                # A folder - place the object inside it
                parent_path, parent_name = dest_path, dest_name
                parent_obj = dest_obj
                
                # Use the source's name for the new object.
                dest_name = src_name
                dest_obj = None
                
                # Construct a destination path.
                dest_path = parent_path + u"/" + src_name
            
            else:
            
                # Not a folder
                
                # dest_path, dest_name and dest_obj are valid values.
                
                # Find the parent object.
                parent_path, parent_name, parent_obj = \
                    self.find_parent_from_path(dest.path())
        
        else:
        
            # Find the parent-to-be from the URL given.
            parent_path, parent_name, parent_obj = \
                self.find_parent_from_path(dest_path)
            
            try:
            
                dest_name = dest_path.split(u"/")[-1]
            
            except IndexError:
            
                self.error(KIO.ERR_COULD_NOT_WRITE, dest_path)
                return
            
            # dest_path is a valid, dest_obj is None
        
        # At this point we should have defined the following:
        #
        # parent_path, parent_name, parent_obj, dest_name, dest_path
        #
        # The following may have been defined:
        #
        # dest_obj
        
        if parent_obj is None:
        
            # No parent to contain a new object.
            self.error(KIO.ERR_COULD_NOT_WRITE, dest_path)
            return
        
        # If a destination object has not yet been identified then we need
        # to determine whether any exist at the path we will be using.
        
        if dest_obj is None:
        
            # Check for existing objects.
            name, dest_obj = self.find_object_in_structure(dest_path)
            
            if (not overwrite) and (dest_obj is not None):
            
                # Can't overwrite existing object.
                self.error(KIO.ERR_COULD_NOT_WRITE, dest_path)
                return
            
            elif overwrite and (dest_obj is not None):
            
                # Overwrite an existing destination object.
                parent_obj.removeChild(dest_obj)
                #parent_obj.unlink()
        
        # Create a deep copy of the node, placing it at the new location
        # then remove the old node.
        obj_copy = src_obj.cloneNode(1)
        
        if unicode(parent_obj.nodeName()) != u"folder" and \
            unicode(parent_obj.nodeName()) != u"xbel":
        
            self.error(KIO.ERR_IS_FILE, parent_path)
            return
        
        # Add the new node to the parent's children.
        parent_obj.appendChild(obj_copy)
        
        # Change the name of the new object.
        self.retitle_object(obj_copy, dest_name)
        
        # Synchronise the filesystem with the bookmarks file.
        self._flush()
        
        self.finished()
    
    #copy = trap(copy)
    
    def del_(self, url, isfile):
    
        #debug("del_")
        self.openConnection()
        
        # Find the path, name and object from the URL given.
        path, name, obj = self.find_object_from_url(url)
        
        #debug("path: %s" % repr(path))
        #debug("name: %s" % repr(name))
        #debug("obj: %s" % repr(obj))
        
        if obj is None:
        
            self.error(KIO.ERR_DOES_NOT_EXIST, path)
            return
        
        elif isfile == 1:
        
            if unicode(obj.nodeName()) != u"bookmark":
            
                self.error(KIO.ERR_IS_DIRECTORY, path)
                return
        
        elif isfile == 0:
        
            if unicode(obj.nodeName()) == u"bookmark":
            
                self.error(KIO.ERR_IS_FILE, path)
                return
        
        # Delete the object found.
        parent_obj = obj.parentNode()
        removed = parent_obj.removeChild(obj)
        
        #debug("remove failed?: %s" % removed.isNull())
        #parent_obj.unlink()
        
        # Synchronise the filesystem with the bookmarks file.
        self._flush()
        
        #debug("self: %s" % repr(self))
        #debug("document: %s" % repr(self.document))
        
        self.finished()
    
    #del_ = trap(del_)
    
    def disconnectSlave(self):
    
        #debug("disconnectSlave")
        return
    
    def dispatchLoop(self):
    
        #debug("dispatchLoop")
        KIO.SlaveBase.dispatchLoop(self)
    
    def strip_text_nodes(self, nodes):
    
        # Remove any Text nodes from the list of nodes given.
        return filter(lambda x: not isinstance(x, minidom.Text), nodes)
    
    def find_node(self, node, name):
    
        child = node.firstChild()
        
        while not child.isNull():
        
            if unicode(child.nodeName()) == name:
            
                return child
            
            child = child.nextSibling()
        
        return None
    
    def find_named_object(self, node, name):
    
        """node = find_named_object(self, nodes, name)
        
        Examine each of the nodes presented, looking for those with
        <title> child nodes and returning a node if one containing the
        appropriate text is found.
        
        If no matching nodes are found then None is returned.
        """
        
        child = node.firstChild()
        
        while not child.isNull():
        
            # Retrieve the title child node, if there is one.
            title = self.find_node(child, u"title")
            
            if title is not None:
            
                text = title.firstChild().nodeValue()
                etext = self.encode_name(text)
                
                if etext == name:
                
                    return child
            
            child = child.nextSibling()
        
        return None
    
    def find_object_in_structure(self, path):
    
        # Descend the directory structure looking for the directory specified.
        try:
        
            # Go to the document element.
            node = self.document.documentElement()
        
        except IndexError:
        
            return None, None
        
        elements = filter(lambda x: x != u'', path.split(u"/"))
        
        level = 1
        
        if elements == []:
        
            return u"", node
        
        for element in elements:
        
            # Remove any text nodes.
            #files = self.strip_text_nodes(node.childNodes)
            
            node = self.find_named_object(node, element)
            
            if node is not None:
            
                name = element
                obj = node
                
                if unicode(obj.nodeName()) != u"folder":
                
                    # This item is not a directory. This only matters if
                    # we have not decended to the lowest level of the
                    # structure required or if we are looking for a directory.
                    if level < len(elements):
                    
                        return None, None
                
                level = level + 1
            
            else:
            
                return None, None
        
        return name, obj
    
    def build_entry(self, obj):
    
        entry = []
        
        # Check the type of object passed.
        if unicode(obj.nodeName()) == u"bookmark":
        
            details = self.read_bookmark_details(obj)
            
            # Decode the title string so that it is presented in a more
            # conventional manner.
            details[u"title"] = self.decode_name(details[u"title"])
            
            output = self.desktop_template % details
            
            length = len(output.encode(details[u"encoding"]))
            mimetype = self.bookmark_mimetype
        
        elif unicode(obj.nodeName()) == u"folder" or \
            unicode(obj.nodeName()) == u"xbel":
        
            length = 0
            mimetype = "inode/directory"
        
        else:
        
            return entry
        
        # Retrieve the title child node, if there is one.
        title = self.find_node(obj, u"title")
        
        if title is not None:
        
            name = self.encode_name(title.firstChild().nodeValue())
        
        elif unicode(obj.nodeName()) == u"xbel":
        
            name = u"/"
        
        else:
        
            return []
        
        atom = KIO.UDSAtom()
        atom.m_uds = KIO.UDS_NAME
        atom.m_str = name
        
        #debug("name: %s" % name)
        entry.append(atom)
        
        atom = KIO.UDSAtom()
        atom.m_uds = KIO.UDS_SIZE
        atom.m_long = length
        
        #debug("length: %i" % length)
        entry.append(atom)
        
        atom = KIO.UDSAtom()
        atom.m_uds = KIO.UDS_MODIFICATION_TIME
        # Number of seconds since the epoch.
        atom.m_long = int(time.time())
        
        entry.append(atom)
        
        atom = KIO.UDSAtom()
        atom.m_uds = KIO.UDS_ACCESS
        # The usual octal permission information (rw-r--r-- in this case).
        atom.m_long = 0644
        
        entry.append(atom)
        
        # If the stat method is implemented then entries _must_ include
        # the UDE_FILE_TYPE atom or the whole system may not work at all.
        
        atom = KIO.UDSAtom()
        atom.m_uds = KIO.UDS_FILE_TYPE
        
        if unicode(obj.nodeName()) == u"bookmark":
        
            atom.m_long = os.path.stat.S_IFREG
        
        else:
        
            atom.m_long = os.path.stat.S_IFDIR
        
        entry.append(atom)
        
        atom = KIO.UDSAtom()
        atom.m_uds = KIO.UDS_MIME_TYPE
        atom.m_str = mimetype
        
        entry.append(atom)
        
        return entry
    
    #build_entry = trap(build_entry)
    
    def get_attr(self, obj):
    
        if obj.hasAttributes() == 0:
        
            return {}
        
        obj_attr = obj.attributes()
        attr = {}
        
        for i in range(0, obj_attr.length()):
        
            attr[unicode(obj_attr.item(i).nodeName())] = \
                unicode(obj_attr.item(i).nodeValue())
        
        return attr
    
    def read_bookmark_details(self, obj):
    
        """details = read_bookmark_details(self, obj)
        
        Read the relevant bookmark attributes and content, returning
        the information in a dictionary suitable for use with an HTML
        page template.
        """
        
        details = self.get_attr(obj)
        
        #debug("%s" % repr(details))
        
        if not details.has_key(u"icon") or details[u"icon"] == u"":
        
            details[u"icon"] = u"www"
        
        # Retrieve the title child node, if there is one.
        title = self.find_node(obj, u"title")
        
        if title is not None:
        
            text = self.encode_name(title.firstChild().nodeValue())
        
        else:
        
            text = u"[No title]"
        
        details[u"title"] = text
        details[u"encoding"] = u"UTF-8"
        
        return details
    
    def encode_name(self, name):
    
        #return unicode(KURL.encode_string_no_slash(name))
        new = u""
        name = unicode(name)
        
        for c in name:
        
            if c == u"/":
                new = new + u"%2f"
            elif c == u"%":
                new = new + u"%25"
            else:
                new = new + c
        
        return new
    
    def decode_name(self, name):
    
        #return unicode(KURL.decode_string(name))
        new = u""
        
        i = 0
        
        while i < len(name):
        
            c = name[i:i+3]
            
            if c == u"%2f":
            
                new = new + u"/"
                i = i + 3
            
            elif c == u"%25":
            
                new = new + u"%"
                i = i + 3
            
            else:
            
                new = new + c[:1]
                i = i + 1
        
        # There should be no trailing % symbols.
        return new
    
    def find_object_from_url(self, url):
    
        #debug("find_object_from_url(%s, %s)" % (repr(self), repr(url)))
        
        # Clean up the URL's path.
        url.cleanPath(1)
        
        # Parse the source URL using the URL's methods.
        host, path = unicode(url.host()), unicode(url.path())
        
        name, obj = self.find_object_in_structure(path)
        
        #debug("find_object_from_url")
        #debug("name: %s" % repr(name))
        #debug("obj: %s" % repr(obj))
        
        if obj is None:
        
            return path, None, None
        
        else:
        
            return path, name, obj
    
    #find_object_from_url = trap(find_object_from_url)
    
    def find_parent_from_path(self, path):
    
        # Split the path into elements and check that the path up to
        # the last element exists.
        elements = unicode(path).split(u"/")
        
        parent = u"/".join(elements[:-1])
        
        name, obj = self.find_object_in_structure(parent)
        
        # Return the path of the parent rather than its name.
        return parent, name, obj
    
    def retitle_object(self, obj, name):
    
        # Change the name of the node title.
        title = self.find_node(obj, u"title")
        
        #debug("retitle: %s %s" % (title.firstChild().nodeValue(), self.decode_name(name)))
        title.firstChild().setNodeValue(self.decode_name(name))

    def create_child(self, node_name, attributes, title, parent):
    
        # Add a new element to the parent node.
        obj = self.document.createElement(node_name)
        
        for aname, avalue in attributes.items():
        
            obj.setAttribute(aname, avalue)
        
        parent.appendChild(obj)
        
        # Give the folder a title element.
        title_node = self.document.createElement(u"title")
        obj.appendChild(title_node)
        
        # Place some text within the title element.
        name = self.document.createTextNode(title)
        title_node.appendChild(name)
    
    #create_child = trap(create_child)
    
    def extract_bookmark_info(self, text):
    
        # Given some text, locate link target and image information
        # within and return it to the caller.
        
        lines = text.split("\n")
        
        d = {}
        
        for line in lines:
        
            at = line.find("=")
            
            if at != -1:
            
                d[line[:at].lower()] = line[at + 1:]
        
        href = d.get("url", None)
        
        if href is not None:
        
            icon = d.get("icon", None)
            title = d.get("title", None)
        
        else:
        
            icon = None
            title = None
        
        return href, icon, title
