#!/usr/bin/env python

"""
kio_package.py

A kioslave for creating XML descriptions of packages.

Copyright (C) 2005 David Boddie <david@boddie.org.uk>

This software is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of
the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public
License along with this library; see the file COPYING
If not, write to the Free Software Foundation, Inc.,
59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

"""

# For debugging purposes, import the sys and traceback modules.

import sys, traceback

# Debugging variable and functions --------------------------------------------

# Debugging flag: set to True to enable debugging.

DEBUG = False
DEBUG_FILE = "/tmp/logs/pydebug.txt"

def debug_file():

    return open(DEBUG_FILE, "a")

def debug(msg):

    if not DEBUG: return
    
    debug_file().write(msg+"\n")
    sys.stderr.write(msg+"\n")

# Create a trace method to catch errors in methods.

def trace(fn, args, file = None):

    debug("tracing %s%s" % (fn.__name__, args))
    try:
    
        result = fn(*args)
        debug("exiting %s with %s\n" % (fn.__name__, repr(result)))
        return result
    
    except:
    
        if file is None:
            file = open(DEBUG_FILE, "a")
        
        traceback.print_exception(
            *(sys.exc_info()), **{"file": file}
            )
    
    debug("exiting %s\n" % fn.__name__)

# The trap function allows functions to be called under conditions where
# any exceptions raised are caught and written to the debug file.

def trap(fn):

    """fn = trap(fn)
    
    Create a function wrapper in which the original function is called
    within a try...except clause for debugging purposes.
    """

    old_fn = fn
    return lambda *args: trace(old_fn, args)

# End of debugging functions --------------------------------------------------

# Import the required Qt and KDE modules.

from qt import QByteArray, QDataStream, QFile, QFileInfo, QString, \
               QStringList, IO_ReadOnly, IO_WriteOnly, SIGNAL
from kio import KMimeType, KIO
from kdecore import KLockFile, KURL, locateLocal

# The os library are required for this IOSlave to function.
import os

# The Qt XML library is used to read and interpret the package file.
import qtxml


# Define a class which will be used to create IOSlave instances.

class SlaveClass(KIO.SlaveBase):

    """SlaveClass(KIO.SlaveBase)
    
    See kdelibs/kio/kio/slavebase.h for virtual functions to override.
    """

    def __init__(self, pool, app):
    
        # We must call the initialisation method of the base class.
        
        KIO.SlaveBase.__init__(self, "package", pool, app)
        
        # Attach the DCOP client object associated with this IOSlave to the
        # DCOP server.
        self.dcopClient().attach()
        
        #debug("dcopClient: %i" % self.dcopClient().isRegistered())
        
        # Initialise some instance variables.
        self.host = ""
        self.document = None
        self.file = None
        self.temp_file = None
        self.lock_file = None
    
    #__init__ = trap(__init__)
    
    def __del__(self):
    
        pass
    
    def setHost(self, host, port, user, passwd):
    
        # This IOSlave does not allow a host to be specified as part of
        # a URL.
        
        if unicode(host) != u"":
        
            #self.closeConnection()
            self.error(KIO.ERR_MALFORMED_URL, host)
            return
    
    #setHost = trap(setHost)
    
    #def setSubURL(self, url):
    #
    #    #debug("setSubURL")
    #    self.finished()
    
    #setSubURL = trap(setSubURL)
    
    def _open(self, file_path):
    
        temp_file = QString(file_path)
        self.temp_file = locateLocal("tmp", temp_file.replace("/", "-"))
        self.lock_file = KLockFile(self.temp_file)
        
        # Wait for the file to be unlocked.
        while self.lock_file.lock() != KLockFile.LockOK:
            pass
        
        f = QFile(file_path)
        if f.open(IO_ReadOnly):
        
            data = f.readAll()
            self.file = file_path
            f.close()
        
        else:
        
            return None
        
        return data
    
    #_open = trap(_open)
    
    def _close(self):
    
        # Delete the temporary file. If something else is waiting for it
        # then its lock will create a new lock file anyway.
        if self.temp_file:
        
            QFile(self.temp_file).remove()
            self.temp_file = None
        
        # This should be unnecessary.
        if self.lock_file:
        
            self.lock_file.unlock()
            self.lock_file = None
    
    #_close = trap(_close)
    
    def _flush(self):
    
        # Internal function for synchronising the filesystem with the
        # package file.
        if self.document is None:
        
            return
        
        try:
        
            f = QFile(self.file)
            if not f.open(IO_WriteOnly):
            
                raise IOError
            
            # Write the XML to the package file.
            data = QByteArray(unicode(self.document.toString()).encode("UTF_8"))
            if f.writeBlock(data) == -1:
            
                raise IOError
            
            f.close()
        
        except IOError:
        
            debug("Failed to write %s" % str(self.file))
            self.error(KIO.ERR_CANNOT_FOR_WRITING, self.file)
            return
        
        # Unset the document attribute to force the bookmarks file to be
        # re-read.
        self.document = None
    
    #_flush = trap(_flush)
    
    def get(self, url):
    
        # Find information about the object from the URL.
        path, name, obj = self.find_object_from_url(url)
        
        if obj is None:
        
            self._close()
            self.error(KIO.ERR_DOES_NOT_EXIST, path)
            return
        
        # The obj variable is a document node. Check whether it is a
        # folder node and return an error if so.
        
        if obj.nodeName() != QString("file"):
        
            self._close()
            self.error(KIO.ERR_IS_DIRECTORY, path)
            return
        
        # Indicate the MIME type of the file.
        self.mimeType(obj.attribute("mimetype"))
        
        obj_url = obj.attribute("url")
        
        if obj_url.isNull():
        
            self._close()
            self.error(KIO.ERR_DOES_NOT_EXIST, obj_url)
            return
        
        else:
        
            debug("url: %s" % str(obj_url))
            url = KURL(obj_url)
            self.redirection(url)
        
        self._close()
        self.finished()
    
    #get = trap(get)
    
    def stat(self, url):
    
        # Find information about the object from the URL.
        path, name, obj = self.find_object_from_url(url)
        
        if obj is None:
        
            url.setProtocol("file")
            self.redirection(url)
            self._close()
            self.finished()
            return
        
        entry = self.build_entry(obj)
        
        if entry != []:
        
            # Return the entry to the application, indicating that it is
            # the only/last entry.
            self.statEntry(entry)
            
            self._close()
            self.finished()
        
        else:
        
            self._close()
            self.error(KIO.ERR_DOES_NOT_EXIST, path)
    
    #stat = trap(stat)
    
    def mimetype(self, url):
    
        # Find information about the object from the URL.
        path, name, obj = self.find_object_from_url(url)
        
        if obj is None:
        
            self._close()
            self.error(KIO.ERR_DOES_NOT_EXIST, path)
            return
        
        self.mimeType(obj.attribute("mimetype"))
        
        self._close()
        self.finished()
    
    #mimetype = trap(mimetype)
    
    def listDir(self, url):
    
        # The "url" argument is a kdecore.KURL object.
        
        # Find information about the object from the URL.
        path, name, obj = self.find_object_from_url(url)
        
        if obj is None:
        
            url.setProtocol("file")
            self.redirection(url)
            self._close()
            self.finished()
            return
        
        if obj.nodeName() == QString("file"):
        
            self._close()
            self.error(KIO.ERR_IS_FILE, path)
            return
        
        # Obtain a list of nodes.
        #files = self.strip_text_nodes(obj.childNodes)
        
        # Return the objects in the list of child nodes to the application.
        child = obj.firstChild()
        
        while not child.isNull():
        
            if child.isElement():
            
                entry = self.build_entry(child.toElement())
                
                if entry != []:
                
                    debug(" entry: %s" % repr(entry))
                    self.listEntry(entry, False)
            
            child = child.nextSibling()
        
        # We have finished listing the contents of a directory.
        self.listEntry([], True)
        
        self._close()
        self.finished()
    
    #listDir = trap(listDir)
    
    def mkdir(self, url, permissions):
    
        # Find the potential parent-to-be of the new directory.
        parent_path, parent_name, obj = self.find_parent_from_url(url)
        
        if obj is None:
        
            self._close()
            self.error(KIO.ERR_DOES_NOT_EXIST, parent_path)
            return
        
        if obj.nodeName() == QString("file"):
        
            self._close()
            self.error(KIO.ERR_IS_FILE, parent_path)
            return
        
        # Find the destination object's name. This is assumed to be the
        # last path element.
        path = url.path(-1)
        
        dest_name = QStringList.split("/", path, False)[-1]
        
        if dest_name.isEmpty():
        
            self._close()
            self.error(KIO.ERR_COULD_NOT_MKDIR, path)
            return
        
        # Create the directory in the node returned.
        
        if self.find_named_object(obj, dest_name) is not None:
        
            self._close()
            self.error(KIO.ERR_COULD_NOT_MKDIR, path)
            return
        
        # Add a new directory element to the parent node.
        attributes = {
            "name": dest_name, "type": "directory",
            "mimetype": "inode/directory"
            }
        
        self.create_child("directory", attributes, obj)
        
        # Synchronise the filesystem with the bookmarks file.
        self._flush()
        self._close()
        
        self.finished()
    
    #mkdir = trap(mkdir)
    
    def rename(self, src, dest, overwrite):
    
        #debug("rename: %s %s" % (src.path(), dest.path()))
        
        # Find the source leafname and object from the URL given.
        src_path, src_name, src_obj = self.find_object_from_url(src)
        
        if src_obj is None:
        
            self._close()
            self.error(KIO.ERR_DOES_NOT_EXIST, src.path(-1))
            return
        
        # Find the destination leafname and object from the URL given.
        dest_path, dest_name, dest_obj = self.find_object_from_url(dest)
        
        #debug("%s %s" % (repr(src_obj), repr(dest_obj)))
        
        if src_obj is dest_obj:
        
            #debug("No change")
            self._close()
            self.finished()
            return
        
        elif dest_obj is not None:
        
            if not overwrite:
            
                self._close()
                self.error(KIO.ERR_CANNOT_RENAME, dest.path(-1))
                return
            
            else:
            
                # Overwrite an existing destination object.
                parent_obj = dest_obj.parentNode()
                parent_obj.removeChild(dest_obj)
        
        else:
        
            dest_path = dest.path(-1)
        
        # Split the source and destination paths into elements and check
        # whether the locations exist within the same folder.
        src_elements = QStringList.split("/", src_path)
        src_name = src_elements[-1]
        dest_elements = QStringList.split("/", dest_path)
        dest_name = dest_elements[-1]
        
        if src_name.isNull() or dest_name.isNull():
        
            self._close()
            self.error(KIO.ERR_CANNOT_RENAME, dest_path)
            return
        
        # Change the title element within the source object. The title
        # element must exist because it was used to locate the object.
        
        src_leading = src_elements[:-1]
        dest_leading = dest_elements[:-1]
        
        if src_leading == dest_leading:
        
            # Change the name of the source object.
            src_obj.setAttribute("name", dest_name)
        
        else:
        
            # The destination URL is elsewhere within the directory
            # structure. Create a deep copy of the node, placing it at
            # the new location then remove the old node.
            
            # Remove the original node from its parent's list of children.
            src_parent = src_obj.parentNode()
            obj_copy = src_parent.removeChild(src_obj)
            
            # Find the parent-to-be of the destination node.
            parent_path, parent_name, parent_obj = \
                self.find_parent_from_url(dest)
            
            if parent_obj is None:
            
                self._close()
                self.error(KIO.ERR_DOES_NOT_EXIST, parent_path)
                return
            
            if parent_obj.nodeName() == QString("file"):
            
                self._close()
                self.error(KIO.ERR_IS_FILE, parent_path)
                return
            
            # Add the new node to the parent's children.
            obj_copy = parent_obj.appendChild(obj_copy)
            
            if src_name != dest_name:
            
                # Change the name of the new object.
                obj_copy.setAttribute("name", dest_name)
        
        # Synchronise the filesystem with the package file.
        self._flush()
        self._close()
        
        self.finished()
    
    #rename = trap(rename)
    
    #def symlink(self, target, dest, overwrite):
    #
    #    debug("symlink")
    #    if self.document is None:
    #    
    #        self.openConnection()
    #    
    #    self.finished()
    
    # Other possible file operations are represented by the following
    # methods which are not implemented.
    
    #def chmod(self, url, permissions):
    #
    #    debug("chmod")
    #    if self.document is None:
    #    
    #        self.openConnection()
    #    
    #    self.finished()
    
    def copy(self, src, dest, permissions, overwrite):
    
        #debug("copy %s to %s" % (str(src.url(0,0)), str(dest.url(0,0))))
        
        if src.protocol() != QString("file"):
        
            # Find the source path, name and object from the URL given.
            src_path, src_name, src_obj = self.find_object_from_url(src)
            
            if src_obj is None:
            
                self._close()
                self.error(KIO.ERR_DOES_NOT_EXIST, src.path(-1))
                return
        
        else:
        
            # Find the source path, name and object from the URL given.
            src_path = src.path(-1)
            src_name = QStringList.split("/", src_path, False)[-1]
            # Define the source object to be None to indicate that a
            # new element has to be created.
            src_obj = None
        
        # Find the destination path, name and object from the URL given.
        dest_path, dest_name, dest_obj = self.find_object_from_url(dest)
        debug("dest_obj: %s" % repr(dest_obj))
        
        if dest_obj is not None:
        
            # Object already exists
            
            if dest_obj.nodeName() != QString("file"):
            
                # A directory - place the object inside it.
                parent_path, parent_name = dest_path, dest_name
                parent_obj = dest_obj
                
                # Use the source's name for the new object.
                dest_name = src_name
                dest_obj = None
                
                # Construct a destination path.
                dest_path = parent_path.append("/").append(src_name)
            
            else:
            
                # Not a folder
                
                # dest_path, dest_name and dest_obj are valid values.
                
                # Find the parent object.
                parent_path, parent_name, parent_obj = \
                    self.find_parent_from_url(dest)
        
        else:
        
            # Find the parent-to-be from the URL given.
            parent_path, parent_name, parent_obj = \
                self.find_parent_from_url(dest)
            
            dest_path = dest.path(-1)
            
            try:
            
                dest_name = QStringList.split("/", dest_path, False)[-1]
            
            except IndexError:
            
                self._close()
                self.error(KIO.ERR_COULD_NOT_WRITE, dest_path)
                return
            
            # dest_path is a valid QString; dest_obj is None
        
        # At this point we should have defined the following:
        #
        # parent_path, parent_name, parent_obj, dest_name, dest_path
        #
        # The following may have been defined:
        #
        # dest_obj
        
        if parent_obj is None:
        
            # No parent to contain a new object.
            self._close()
            self.error(KIO.ERR_COULD_NOT_WRITE, dest_path)
            return
        
        elif parent_obj.nodeName() == QString("file"):
        
            self._close()
            self.error(KIO.ERR_IS_FILE, parent_path)
            return
        
        # If a destination object has not yet been identified then we need
        # to determine whether any exist at the path we will be using.
        
        if dest_obj is None:
        
            # Check for existing objects.
            name, dest_obj = self.find_object_in_structure(dest_path)
        
        if dest_obj is not None:
        
            if not overwrite:
        
                # Can't overwrite existing object.
                self._close()
                self.error(KIO.ERR_COULD_NOT_WRITE, dest_path)
                return
            
            else:
            
                # Overwrite an existing destination object.
                parent_obj.removeChild(dest_obj)
        
        if src_obj:
        
            # Create a deep copy of the node, placing it at the new location
            # then remove the old node.
            obj_copy = src_obj.cloneNode(1).toElement()
            
            # Add the new node to the parent's children.
            parent_obj.appendChild(obj_copy)
            
            # Change the name of the new object.
            obj_copy.setAttribute("name", dest_name)
        
        else:
        
            # Add a new element to the parent element.
            attributes = {
                "name": dest_name,
                "mimetype": KMimeType.findByURL(src).name(),
                "url": src.url(0, 0)
                }
            
            self.create_child("file", attributes, parent_obj)
        
        # Synchronise the filesystem with the package file.
        self._flush()
        self._close()
        
        self.finished()
    
    #copy = trap(copy)
    
    def del_(self, url, isfile):
    
        #debug("del_")
        
        # Find the path, name and object from the URL given.
        path, name, obj = self.find_object_from_url(url)
        
        if obj is None:
        
            self._close()
            self.error(KIO.ERR_DOES_NOT_EXIST, path)
            return
        
        elif isfile == 1:
        
            if obj.nodeName() != QString("file"):
            
                self._close()
                self.error(KIO.ERR_IS_DIRECTORY, path)
                return
        
        elif isfile == 0:
        
            if obj.nodeName() == QString("file"):
            
                self._close()
                self.error(KIO.ERR_IS_FILE, path)
                return
        
        # Delete the object found.
        parent_obj = obj.parentNode()
        removed = parent_obj.removeChild(obj)
        
        # Synchronise the filesystem with the bookmarks file.
        self._flush()
        self._close()
        
        self.finished()
    
    #del_ = trap(del_)
    
    def disconnectSlave(self):
    
        #debug("disconnectSlave")
        return
    
    def dispatchLoop(self):
    
        #debug("dispatchLoop")
        KIO.SlaveBase.dispatchLoop(self)
    
    def build_entry(self, obj):
    
        entry = []
        path = obj.attribute("url")
        info = QFileInfo(path)
        
        # Check the type of object passed.
        if obj.nodeName() == QString("file"):
        
            name = self.encode_name(obj.attribute("name"))
            permissions = 0644
            filetype = os.path.stat.S_IFREG
            mimetype = obj.attribute("mimetype")
        
        elif obj.nodeName() == QString("directory"):
        
            name = self.encode_name(obj.attribute("name"))
            permissions = 0544
            filetype = os.path.stat.S_IFDIR
            mimetype = obj.attribute("mimetype")
        
        elif obj.nodeName() == QString("package"):
        
            name = QString("/")
            permissions = 0544
            filetype = os.path.stat.S_IFDIR
            mimetype = "inode/directory"
        
        else:
        
            return entry
        
        atom = KIO.UDSAtom()
        atom.m_uds = KIO.UDS_NAME
        atom.m_str = name
        
        #debug("name: %s" % name)
        entry.append(atom)
        
        atom = KIO.UDSAtom()
        atom.m_uds = KIO.UDS_SIZE
        atom.m_long = info.size()
        
        #debug("length: %i" % length)
        entry.append(atom)
        
        atom = KIO.UDSAtom()
        atom.m_uds = KIO.UDS_MODIFICATION_TIME
        # Number of seconds since the epoch.
        atom.m_long = info.lastModified().toTime_t()
        
        entry.append(atom)
        
        atom = KIO.UDSAtom()
        atom.m_uds = KIO.UDS_ACCESS
        # The usual octal permission information (rw-r--r-- in this case).
        atom.m_long = permissions
        
        entry.append(atom)
        
        # If the stat method is implemented then entries _must_ include
        # the UDE_FILE_TYPE atom or the whole system may not work at all.
        
        atom = KIO.UDSAtom()
        atom.m_uds = KIO.UDS_FILE_TYPE
        atom.m_long = filetype
        
        entry.append(atom)
        
        atom = KIO.UDSAtom()
        atom.m_uds = KIO.UDS_MIME_TYPE
        atom.m_str = mimetype
        
        entry.append(atom)
        
        return entry
    
    #build_entry = trap(build_entry)
    
    def find_node(self, node, name):
    
        child = node.firstChild()
        
        while not child.isNull():
        
            if unicode(child.nodeName()) == name:
            
                return child
            
            child = child.nextSibling()
        
        return None
    
    def find_named_object(self, node, name):
    
        """node = find_named_object(self, node, name)
        
        Examine each of the child nodes of the specified node, looking for
        those with a matching "name" attribute, returning the first node
        that matches.
        
        If no matching nodes are found then None is returned.
        """
        
        child = node.firstChild()
        
        while not child.isNull():
        
            # Retrieve the name attribute, if there is one.
            if child.isElement():
            
                child = child.toElement()
                elementName = child.attribute("name")
                if elementName == name:
                
                    return child
            
            child = child.nextSibling()
        
        return None
    
    def find_object_in_structure(self, path):
    
        # Descend the directory structure looking for the directory specified.
        try:
        
            # Go to the document element.
            node = self.document.documentElement()
        
        except IndexError:
        
            return None, None
        
        elements = QStringList.split("/", path, False)
        
        level = 1
        
        if elements.count() == 0:
        
            return QString("/"), node
        
        elif elements.count() == 1 and elements[0].isEmpty():
        
            return QString("/"), node
        
        for element in elements:
        
            # Remove any text nodes.
            #files = self.strip_text_nodes(node.childNodes)
            
            node = self.find_named_object(node, element)
            
            if node is not None:
            
                name = element
                obj = node
                
                if obj.nodeName() == QString("file"):
                
                    # This item is a file. This only matters if we have not
                    # decended to the lowest level of the structure required
                    # or if we are looking for a directory.
                    if level < len(elements):
                    
                        return None, None
                
                level = level + 1
            
            else:
            
                return None, None
        
        if obj.isElement():
        
            return name, obj.toElement()
        
        else:
        
            return None, None
    
    #find_object_in_structure = trap(find_object_in_structure)
    
    def find_object_from_url(self, url):
    
        #debug("find_object_from_url(%s, %s)" % (repr(self), repr(url)))
        
        # If we have cached a URL for the disc image then check whether
        # this URL refers to an item beneath it. This implies that the
        # cached URL is a substring of the one given.
        
        # Clean up the URL's path.
        url.cleanPath(1)
        
        file_path = None
        package_path = None
        
        #if self.document:
        #
        #    path = url.path(-1)
        #    at = path.find(self.file)
        #    
        #    if at == 0:
        #    
        #        file_path = self.file
        #        package_path = path.mid(self.file.length())
        
        # If file_path is undefined then either a document has not been
        # opened, or the path specified refers to another document. In either
        # case we need to open a document.
        
        if file_path is None:
        
            # Split the path (minus any trailing /) into elements.
            elements = QStringList.split("/", url.path(-1), True)
            
            path_elements = QStringList()
            path_elements.append(elements[0])
            del elements[0]
            
            while elements.count() > 0:
            
                path_elements.append(elements[0])
                del elements[0]
                
                info = QFileInfo(QStringList.join(path_elements, "/"))
                
                if info.isFile():
                
                    break
                
                elif elements.count() == 0 and info.isDir():
                
                    return None, None, None
                
                elif not info.exists():
                
                    return None, None, None
            
            file_path = QStringList.join(path_elements, "/")
            package_path = QStringList.join(elements, "/")
            
            data = self._open(file_path)
            
            if data is None:
            
                return url.path(-1), None, None
            
            # Convert the XML to a DOM representation.
            self.document = qtxml.QDomDocument()
            
            # Read the file into the document, converting the XML into a
            # DOM representation.
            
            result, errorMsg, errorLine, errorColumn = self.document.setContent(data)
            
            if result == 0:
            
                return url.path(-1), None, None
        
        # Find the referenced object in the package structure.
        name, obj = self.find_object_in_structure(package_path)
        
        if obj is None:
        
            return url.path(-1), None, None
        
        else:
        
            return url.path(-1), name, obj
    
    #find_object_from_url = trap(find_object_from_url)
    
    def find_parent_from_url(self, url):
    
        # Clean up the URL's path.
        url.cleanPath(1)
        url = url.upURL()
        
        # Find the parent.
        return self.find_object_from_url(url)
    
    def get_attr(self, obj):
    
        if obj.hasAttributes() == 0:
        
            return {}
        
        obj_attr = obj.attributes()
        attr = {}
        
        for i in range(0, obj_attr.length()):
        
            attr[unicode(obj_attr.item(i).nodeName())] = \
                obj_attr.item(i).nodeValue()
        
        return attr
    
    def encode_name(self, name):
    
        #return unicode(KURL.encode_string_no_slash(name))
        new = u""
        name = unicode(name)
        
        for c in name:
        
            if c == u"/":
                new = new + u"%2f"
            elif c == u"%":
                new = new + u"%25"
            else:
                new = new + c
        
        return new
    
    def decode_name(self, name):
    
        #return unicode(KURL.decode_string(name))
        new = u""
        
        i = 0
        
        while i < len(name):
        
            c = name[i:i+3]
            
            if c == u"%2f":
            
                new = new + u"/"
                i = i + 3
            
            elif c == u"%25":
            
                new = new + u"%"
                i = i + 3
            
            else:
            
                new = new + c[:1]
                i = i + 1
        
        # There should be no trailing % symbols.
        return new
    
    def create_child(self, node_name, attributes, parent):
    
        # Add a new element to the parent node.
        obj = self.document.createElement(node_name)
        
        for aname, avalue in attributes.items():
        
            obj.setAttribute(aname, avalue)
        
        parent.appendChild(obj)
    
    #create_child = trap(create_child)
