/*
 * kio_launcher.cpp
 * 
 * Launch kioslaves written in Python using an embedded Python interpreter.
 * 
 * Copyright (C) 2003-2005 Jim Bublitz, David Boddie <david@boddie.org.uk>
 * 
 * This software is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this library; see the file COPYING
 * If not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include <Python.h>
#include <kinstance.h>
#include "kio_launcher.h"

#define debug 0

#if debug == 1
FILE *d = fopen("/tmp/kiodebug.txt", "a");
#define pyDebug(code) \
    PyRun_SimpleString("open('/tmp/logs/kio-debug.txt', 'a').write("code"+'\\n')\n");
#endif

PyObject* call_object(PyObject *callable, PyObject *args)
{
    PyObject *result = NULL, *pArgs;
    
    if (callable == NULL)
    {
        return NULL;
    }
    
    if (PyCallable_Check(callable))
    {
        if (args == NULL)
            pArgs = PyTuple_New(0);
        else
            pArgs = args;
        
        result = PyObject_CallObject(callable, pArgs);
        
        /* If the arguments were created in this function then decrease
           their reference count. */
        if (args == NULL)
            Py_XDECREF(pArgs);
            /* pDict and pFunc are borrowed and must not be Py_DECREF-ed */
        
        if (result == NULL)
        {
            #if debug == 1
            PyErr_Print();
            #else
            PyErr_Clear();
            #endif
        }
    }
    
    return result;
}

PyObject* call_method(PyObject *callable, const char *method)
{
    PyObject *result = NULL;
    
    if (callable == NULL)
    {
        return NULL;
    }
    
    result = PyObject_CallMethod(callable, (char *)method, NULL);

    if (result == NULL)
    {
        #if debug == 1
        PyErr_Print();
        #else
        PyErr_Clear();
        #endif
    }

    return result;
}

extern "C"
{
    /*
        The main function initializes the Python interpreter, if not already
        running, imports the Python module containing the kioslave class to
        use, and performs the equivalent of the following Python code:

        def SlaveFactory(pool, app):

            slave = SlaveClass(pool, app)
            slave.dispatchLoop()
    */

    int kdemain( int argc, char **argv )
    {
        PyObject *pModule, *pName, *pDict;
        KInstance slave(MODULE_NAME);
        
        Py_SetProgramName(argv[0]);
        Py_Initialize();
        
        //PyEval_InitThreads();
        PySys_SetArgv(1, argv);
        
        PyRun_SimpleString("import sys\n");
        PyRun_SimpleString("sys.path.append('"MODULE_DIR"')\n");
        
        pName = PyString_FromString(MODULE_NAME);
        
        pModule = PyImport_Import(pName);
        
        Py_XDECREF(pName);
        
        if (pModule == NULL)
        {
            return 1;
        }
        else
        {
            PyObject *pClass, *pMethod, *pArgs, *pArg1, *pArg2;
            
            pDict = PyModule_GetDict(pModule);
            /* pDict is a borrowed reference */
            
            pClass = PyDict_GetItemString(pDict, FACTORY);
            
            if (pClass == NULL)
            {
                return 1;
            }
            else
            {
                pArgs = PyTuple_New(2);
                
                pArg1 = PyString_FromString(argv[2]);
                pArg2 = PyString_FromString(argv[3]);
                
                PyTuple_SetItem(pArgs, 0, pArg1);
                PyTuple_SetItem(pArgs, 1, pArg2);
                
                PyObject *pObject = call_object(pClass, pArgs);
                
                Py_XDECREF(pClass);
                Py_XDECREF(pArgs);
                
                call_method(pObject, "dispatchLoop");
                
                /* Some time later... */
                
                Py_XDECREF(pObject);
            }
            
            Py_XDECREF(pModule);
        }
        
        Py_Finalize();
        return 0;
    }
}
