#!/usr/bin/env python

"""
kio_python.py

A kioslave for displaying information about a running Python interpreter.

Copyright (C) 2005 David Boddie <david@boddie.org.uk>

This software is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of
the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public
License along with this library; see the file COPYING
If not, write to the Free Software Foundation, Inc.,
59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

"""

# For debugging purposes, import the sys and traceback modules.
# The os and site modules are required for this IOSlave to function.

import os, site, sys, traceback

# Debugging variable and functions --------------------------------------------

# Debugging flag: set to True to enable debugging.

DEBUG = False
DEBUG_FILE = "/tmp/logs/pydebug.txt"

def debug_file():

    return open(DEBUG_FILE, "a")

def debug(msg):

    if not DEBUG: return
    
    debug_file().write(msg+"\n")
    sys.stderr.write(msg+"\n")

# Create a trace method to catch errors in methods.

def trace(fn, args, file = None):

    debug("tracing %s%s" % (fn.__name__, args))
    try:
    
        result = fn(*args)
        debug("exiting %s with %s\n" % (fn.__name__, repr(result)))
        return result
    
    except:
    
        if file is None:
            file = open(DEBUG_FILE, "a")
        
        traceback.print_exception(
            *(sys.exc_info()), **{"file": file}
            )
    
    debug("exiting %s\n" % fn.__name__)

# The trap function allows functions to be called under conditions where
# any exceptions raised are caught and written to the debug file.

def trap(fn):

    """fn = trap(fn)
    
    Create a function wrapper in which the original function is called
    within a try...except clause for debugging purposes.
    """

    old_fn = fn
    return lambda *args: trace(old_fn, args)

# End of debugging functions --------------------------------------------------

# Import the required Qt and KDE modules.

from qt import QByteArray, QDataStream, QFile, QFileInfo, QString, \
               QStringList, IO_ReadOnly, IO_WriteOnly, SIGNAL
from kio import KIO
from kdecore import KURL

import inspect

# Define an exception to be used to indicate failure in the kioslave.

class OperationFailed(Exception):

    pass

# Define a class which will be used to create IOSlave instances.

class SlaveClass(KIO.SlaveBase):

    """SlaveClass(KIO.SlaveBase)
    
    See kdelibs/kio/kio/slavebase.h for virtual functions to override.
    """

    def __init__(self, pool, app):
    
        # We must call the initialisation method of the base class.
        KIO.SlaveBase.__init__(self, "python", pool, app)
        self.encoding = "UTF-8"
    
    def setHost(self, host, port, user, passwd):
    
        # This IOSlave does not allow a host to be specified as part of
        # a URL.
        
        if unicode(host) != u"":
        
            self.error(KIO.ERR_MALFORMED_URL, host)
            return
    
    def stat(self, url):
    
        # Find information about the object from the URL.
        try:
        
            name, obj = self.find_object_from_url(url)
        
        except OperationFailed:
        
            self.error(KIO.ERR_DOES_NOT_EXIST, url.path())
            return
        
        entry = self.build_entry(name, obj)
        
        if entry != []:
        
            # Return the entry to the application, indicating that it is
            # the only/last entry.
            self.statEntry(entry)
            self.finished()
        
        else:
        
            self.error(KIO.ERR_DOES_NOT_EXIST, url.path())
    
    def mimetype(self, url):
    
        # Find information about the object from the URL.
        try:
        
            name, obj = self.find_object_from_url(url)
        
        except OperationFailed:
        
            self.error(KIO.ERR_DOES_NOT_EXIST, url.path())
            return
        
        if self.is_container(obj):
        
            self.mimeType("inode/directory")
        
        else:
        
            self.mimeType("text/html")
        
        self.finished()
    
    def get(self, url):
    
        # Find information about the object from the URL.
        try:
        
            name, obj = self.find_object_from_url(url)
        
        except OperationFailed:
        
            self.error(KIO.ERR_DOES_NOT_EXIST, url.path())
            return
        
        # Check whether the object is a container and return an error if so.
        
        if self.is_container(obj):
        
            self.error(KIO.ERR_IS_DIRECTORY, url.path())
            return
        
        # Indicate the MIME type of the file.
        self.mimeType("text/html")
        
        self.data((
            u"<html>\n"
            u"<head>\n"
            u' <meta http-equiv="Content-Type" content="text/html; charset=%s">\n'
            u' <title>%s</title>\n'
            u"</head>\n"
            u"<body>\n"
            u"<h1>%s</h1>\n" % (self.html_repr(self.encoding),
                                self.html_repr(name),
                                self.html_repr(name))
            ).encode(self.encoding))
        
        source_file = None
        source = None
        
        try:
        
            source_file = inspect.getabsfile(obj)
            source_lines, source_start = inspect.getsourcelines(obj)
            source = u"".join(source_lines)
        
        except (TypeError, IOError):
        
            source = repr(obj)
            source_start = None
        
        if source_file is not None:
        
            self.data((
                u'Defined in <strong><a href="file:%s">%s</a></strong>' % (
                    self.html_repr(source_file), self.html_repr(source_file))
                ).encode(self.encoding))
            
            if source_start is not None:
            
                self.data((
                    u" (line %s)" % self.html_repr(source_start)
                    ).encode(self.encoding))
            
            self.data(u"\n".encode(self.encoding))
        
        self.data((
            u"<pre>\n"
            u"%s\n"
            u"</pre>\n" % self.html_repr(source)
            ).encode(self.encoding))
        
        self.data((
            u"</body>\n"
            u"</html>\n"
            ).encode(self.encoding))
        
        self.data(QByteArray())
        
        self.finished()
    
    def listDir(self, url):
    
        # The "url" argument is a kdecore.KURL object.
        
        # Find information about the object from the URL.
        try:
        
            name, obj = self.find_object_from_url(url)
        
        except OperationFailed:
        
            self.error(KIO.ERR_DOES_NOT_EXIST, url.path())
            return
        
        if not self.is_container(obj):
        
            self.error(KIO.ERR_IS_FILE, url.path())
            return
        
        # Obtain a list of objects and return it to the application.
        for name in obj.__dict__.keys():
        
            entry = self.build_entry(name, obj.__dict__[name])
            
            if entry != []:
            
                self.listEntry(entry, False)
        
        # We have finished listing the contents of a directory.
        self.listEntry([], True)
        self.finished()
    
    def disconnectSlave(self):
    
        return
    
    def dispatchLoop(self):
    
        KIO.SlaveBase.dispatchLoop(self)
    
    def build_entry(self, name, obj):
    
        entry = []
        name = self.encode_name(name)
        length = len(repr(obj).encode(self.encoding))
        
        # Check the type of object passed.
        if self.is_container(obj):
        
            permissions = 0544
            filetype = os.path.stat.S_IFDIR
            mimetype = "inode/directory"
        
        else:
        
            permissions = 0644
            filetype = os.path.stat.S_IFREG
            mimetype = "text/html"
        
        atom = KIO.UDSAtom()
        atom.m_uds = KIO.UDS_NAME
        atom.m_str = name
        
        entry.append(atom)
        
        atom = KIO.UDSAtom()
        atom.m_uds = KIO.UDS_SIZE
        atom.m_long = length
        
        entry.append(atom)
        
        atom = KIO.UDSAtom()
        atom.m_uds = KIO.UDS_ACCESS
        # The usual octal permission information (rw-r--r-- in this case).
        atom.m_long = permissions
        
        entry.append(atom)
        
        # If the stat method is implemented then entries _must_ include
        # the UDE_FILE_TYPE atom or the whole system may not work at all.
        
        atom = KIO.UDSAtom()
        atom.m_uds = KIO.UDS_FILE_TYPE
        atom.m_long = filetype
        
        entry.append(atom)
        
        atom = KIO.UDSAtom()
        atom.m_uds = KIO.UDS_MIME_TYPE
        atom.m_str = mimetype
        
        entry.append(atom)
        
        return entry
    
    def find_object_from_url(self, url):
    
        # Clean up the URL's path.
        url.cleanPath(1)
        
        # Split the path (minus any trailing /) into elements.
        path = unicode(url.path(-1))
        
        elements = filter(lambda element: element != u"", path.split(u"/"))
        
        name = u"kio_python"
        obj = sys.modules["kio_python"]
        
        while elements != []:
        
            name = elements.pop(0)
            
            if name in dir(obj):
            
                obj = getattr(obj, name)
                
                if not self.is_container(obj) and elements != []:
                
                    raise OperationFailed
            
            else:
            
                raise OperationFailed
        
        return name, obj
    
    def is_container(self, obj):
    
        return hasattr(obj, "__dict__") and obj.__dict__.keys() != []
    
    def encode_name(self, name):
    
        #return unicode(KURL.encode_string_no_slash(name))
        new = u""
        name = unicode(name)
        
        for c in name:
        
            if c == u"/":
                new = new + u"%2f"
            elif c == u"%":
                new = new + u"%25"
            else:
                new = new + c
        
        return new
    
    def html_repr(self, s):
    
        new = []
        
        for c in unicode(s):
        
            if c == u"&":
                new.append(u"&amp;")
            elif c == u"<":
                new.append(u"&lt;")
            elif c == u">":
                new.append(u"&gt;")
            else:
                new.append(c)
        
        return u"".join(new)
