#!/usr/bin/env python

"""
kio_sprites.py

IOSlave for Acorn Spritefiles.
Use sprites:<URL> to view the contents of an Acorn Spritefile at <URL>.

Copyright (C) 2003-2005 David Boddie <david@boddie.org.uk>

This software is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of
the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public
License along with this library; see the file COPYING
If not, write to the Free Software Foundation, Inc.,
59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

"""

from qt import QString, QByteArray, QDataStream, IO_ReadOnly
from kio import KIO
from kdecore import KURL
import os, sys, traceback, time, types, urllib2

# Import the Python Imaging Library and the spritefile module.

import Image, spritefile

try:

    from cStringIO import StringIO

except ImportError:

    from StringIO import StringIO

# Debugging variable and functions --------------------------------------------

# Debugging flag: set to True to enable debugging.

DEBUG = False

def debug(msg):

    if not DEBUG: return
    
    sys.stderr.write(msg+"\n")


# Create a trace method to catch errors in methods.

def trace(fn, args):

    debug("tracing %s%s" % (fn.__name__, args))
    try:
    
        fn(*args)
    
    except:
    
        traceback.print_exception(
            *(sys.exc_info()), **{"file": sys.stderr}
            )
    
    debug("exiting %s" % fn.__name__)

# The trap function allows functions to be called under conditions where
# any exceptions raised are caught and written to the debug file.

def trap(fn):

    """fn = trap(fn)
    
    Create a function wrapper in which the original function is called
    within a try...except clause for debugging purposes.
    """

    old_fn = fn
    return lambda *args: trace(old_fn, args)

# End of debugging functions --------------------------------------------------


# Define a class which will be used to create IOSlave instances.

class SlaveClass(KIO.SlaveBase):

    """SlaveClass(KIO.SlaveBase)
    
    See kdelibs/kio/kio/slavebase.h for virtual functions to override.
    """
    
    # Define the MIME type and image format to supply to the Python Imaging
    # Library.
    
    image_mimetype = "image/png"
    image_format = "png"
    
    def __init__(self, pool, app):
    
        # We must call the initialisation method of the base class.
        
        KIO.SlaveBase.__init__(self, "sprites", pool, app)
        
        # Initialise various instance variables.
        
        self.host = ""
        self.spritefile = None
        self.url = None
    
    def __del__(self):
    
        pass
    
    def setSubURL(self, url):
    
        #debug("setSubURL")
        self.finished()
    
    #setSubURL = trap(setSubURL)
    
    def parse_url(self, url):
    
        #debug("parse URL: %s" % str(url.url(0,0)))
        
        # Treat the path as a URL in itself and use that to read the
        # spritefile.
        
        # Abandoned KIO job creation.
        #file_url = KURL(url.path())
        #
        #debug("get URL: %s" % url.path())
        #
        ## Open the URL using a KIO Job.
        #args1, args2 = QByteArray(""), QByteArray("")
        #
        #job = KIO.TransferJob(
        #    file_url, KIO.CMD_GET, args1, args2, 0
        #    )
        #
        #debug("created job")
        #
        #QObject.connect(job, SIGNAL("result(KIO::Job*)"), self.slot_get)
        #QObject.connect(
        #    job, SIGNAL("data(KIO::Job*, const QByteArray&)"), self.slot_data
        #    )
        #
        #debug("about to check")
        #
        #self._pending = 1
        #
        #while self._pending == 1:
        #
        #    pass
        
        file_url = unicode(url.path())
        
        if file_url.find(u":") == -1:
        
            file_url = u"file:" + file_url
        
        # The characters following the last separator in the URL correspond
        # to the sprite name.
        at = file_url.rfind(u"/")
        
        if at != -1:
        
            # Convert the name to string form as this is used for keys into
            # the sprites dictionary.
            use_url = file_url[:at]
            name = str(self.decode_name(file_url[at+1:]))
        
        else:
        
            use_url = file_url
            name = ""
        
        
        # If the URL to use is already cached then return the name
        # to the caller.
        if use_url == self.url:
        
            return name
        
        # Use the first part of the split URL to access the file.
        
        try:
        
            data = urllib2.urlopen(use_url).read()
            
            file = StringIO(data)
            file.seek(0)
            
            self.spritefile = spritefile.spritefile(file)
            
            # Close the StringIO object.
            file.close()
            
            # Cache the URL used.
            self.url = use_url
            
            return name
        
        except:
        
            pass
        
        # If the substitute URL is already cached then return the name
        # to the caller.
        if file_url == self.url:
        
            return ""
        
        # Try the original URL without a name.
        
        try:
        
            data = urllib2.urlopen(file_url).read()
            
            file = StringIO(data)
            file.seek(0)
            
            self.spritefile = spritefile.spritefile(file)
            
            # Close the StringIO object.
            file.close()
            
            # Cache the URL used.
            self.url = file_url
            
            name = ""
        
        except:
        
            self.spritefile = None
            name = None
        
        #debug("spritefile: %s name: %s" % (self.spritefile, name))
        
        return name
    
#    def slot_get(self, job):
#    
#        if job.error():
#        
#            self.spritefile = None
#        
#        self._pending = 0
#    
#    def slot_data(self, job, bytearray):
#    
#        data = bytearray.readBytes()
#        debug("data: %s" % repr(data))
#        
#        io = StringIO(data)
#        
#        self.spritefile = spritefile.spritefile(io)
#        self._pending = 0
    
    def get(self, url):
    
        name = self.parse_url(url)
        
        if name is None:
        
            self.error(KIO.ERR_DOES_NOT_EXIST, url.path())
            return
        
        # Find the sprite within the file.
        if not self.spritefile.sprites.has_key(name):
        
            self.error(KIO.ERR_DOES_NOT_EXIST, name)
            return
        
        sprite = self.spritefile.sprites[name]
        
        self.mimeType(self.image_mimetype)
        
        # Create an Image object.
        image = self.sprite_to_image(sprite)
        
        file = StringIO()
        image.save(file, self.image_format)
        file.seek(0, 0)
        
        output = file.read()
        file.close()
        
        #debug("total size: %s" % repr(len(output)))
        
        #self.totalSize(long(len(output)))
        
        self.data(QByteArray(output))
        
        # The end of the data string.
        self.data(QByteArray())
        
        self.finished()
    
    #get = trap(get)
    
    def stat(self, url):
    
        #debug("stat")
        
        name = self.parse_url(url)
        
        #debug("name: %s" % repr(name))
        
        if name is None:
        
            # Try redirecting to the protocol contained in the path.
            redir_url = KURL(url.path())
            self.redirection(redir_url)
            self.finished()
            #self.error(KIO.ERR_DOES_NOT_EXIST, url.path())
            return
        
        # Find the sprite within the file.
        if name != "" and not self.spritefile.sprites.has_key(name):
        
            self.error(KIO.ERR_DOES_NOT_EXIST, name)
            return
        
        if name != "":
        
            sprite = self.spritefile.sprites[name]
            
            entry = self.build_entry(self.encode_name(unicode(name)), sprite)
        
        else:
        
            entry = self.build_entry(u"/", None)
        
        if entry != []:
        
            # Return the entry to the application, indicating that it is
            # the only/last entry.
            self.statEntry(entry)
            
            self.finished()
        
        else:
        
            self.error(KIO.ERR_DOES_NOT_EXIST, name)
    
    #stat = trap(stat)
    
    def mimetype(self, url):
    
        #debug("mimetype: %s" % unicode(url))
        
        name = self.parse_url(url)
        
        if name is None:
        
            self.error(KIO.ERR_DOES_NOT_EXIST, url.path())
            return
        
        # Return the default image MIME type.
        self.mimeType(self.image_mimetype)
        
        self.finished()
    
    #mimetype = trap(mimetype)
    
    def listDir(self, url):
    
        # The "url" argument is a kdecore.KURL object.
        
        #debug("listDir: %s" % str(url.url(0,0)))
        
        name = self.parse_url(url)
        
        if name is None:
        
            # Try redirecting to the protocol contained in the path.
            redir_url = KURL(url.path())
            self.redirection(redir_url)
            self.finished()
            #self.error(KIO.ERR_DOES_NOT_EXIST, url.path())
            return
        
        if self.spritefile.sprites.has_key(name):
        
            self.error(KIO.ERR_IS_FILE, name)
            return
        
        elif name != "":
        
            self.error(KIO.ERR_DOES_NOT_EXIST, name)
            return
        
        # Obtain a list of sprites.
        files = self.spritefile.sprites.keys()
        
        # Return the objects in the list to the application.
        
        for name in files:
        
            #debug("name: %s" % name)
            #debug("attr: %s" % repr(attributes))
            
            entry = self.build_entry(
                self.encode_name(unicode(name)), self.spritefile.sprites[name]
                )
            
            if entry != []:
            
                self.listEntry(entry, 0)
        
        # We have finished listing the contents of a directory.
        self.listEntry([], 1)
        
        self.finished()
    
    #listDir = trap(listDir)
    
    def disconnectSlave(self):
    
        #debug("disconnectSlave")
        pass
    
    def dispatchLoop(self):
    
        #debug("dispatchLoop")
        KIO.SlaveBase.dispatchLoop(self)
    
    def build_entry(self, name, sprite):
    
        entry = []
        
        # Check the type of object passed.
        if sprite is not None:
        
            image = Image.fromstring(
                sprite['mode'], (sprite['width'], sprite['height']),
                sprite['image']
                )
            
            file = StringIO()
            image.save(file, self.image_format)
            file.seek(0, 2)
            
            length = file.tell()
            file.close()
            
            mimetype = self.image_mimetype
        
        else:
        
            length = 0
            mimetype = "inode/directory"
        
        #debug("name: %s" % name)
        
        atom = KIO.UDSAtom()
        atom.m_uds = KIO.UDS_NAME
        atom.m_str = name
        
        entry.append(atom)
        
        atom = KIO.UDSAtom()
        atom.m_uds = KIO.UDS_SIZE
        atom.m_long = length
        
        entry.append(atom)
        
        atom = KIO.UDSAtom()
        atom.m_uds = KIO.UDS_MODIFICATION_TIME
        # Number of seconds since the epoch.
        atom.m_long = int(time.time())
        
        entry.append(atom)
        
        atom = KIO.UDSAtom()
        atom.m_uds = KIO.UDS_ACCESS
        # The usual octal permission information (rw-r--r-- in this case).
        atom.m_long = 0644
        
        entry.append(atom)
        
        # If the stat method is implemented then entries _must_ include
        # the UDE_FILE_TYPE atom or the whole system may not work at all.
        
        atom = KIO.UDSAtom()
        atom.m_uds = KIO.UDS_FILE_TYPE
        
        if sprite is not None:
        
            atom.m_long = os.path.stat.S_IFREG
        
        else:
        
            atom.m_long = os.path.stat.S_IFDIR
        
        entry.append(atom)
        
        atom = KIO.UDSAtom()
        atom.m_uds = KIO.UDS_MIME_TYPE
        atom.m_str = mimetype
        
        entry.append(atom)
        
        return entry
    
    def encode_name(self, name):
    
        return unicode(KURL.encode_string_no_slash(name))
    
    def decode_name(self, name):
    
        return unicode(KURL.decode_string(name))
    
    def sprite_to_image(self, sprite):
    
        try:
        
            image = Image.fromstring(
                sprite['mode'], (sprite['width'], sprite['height']),
                sprite['image']
                )
        
        except KeyError:
        
            #debug("failed to convert %s" % name)
            self.error(KIO.ERR_DOES_NOT_EXIST, name)
            return
        
        if sprite["dpi x"] != sprite["dpi y"]:
        
            # If the horizontal and vertical resolutions of the sprite are
            # not equal then rescale the image.
            if sprite["dpi x"] > sprite["dpi y"]:
            
                # There were fewer pixels per vertical unit of length so
                # increase the number of rows accordingly. This does not
                # try to find a new height which is an integer number of
                # pixels.
                yscale = float(sprite["dpi x"]) / sprite["dpi y"]
                image = image.resize( (image.size[0], image.size[1] * yscale) )
            
            else:
            
                # There were fewer pixels per vertical unit of length so
                # increase the number of rows accordingly. This does not
                # try to find a new height which is an integer number of
                # pixels.
                xscale = float(sprite["dpi x"]) / sprite["dpi y"]
                image = image.resize( (image.size[0] * xscale, image.size[1]) )
        
        return image
